import dpdata
import numpy as np


def get_amber_fp(cutoff: float,
                 target: str = ":1",
                 out: str = None,
            ) -> dpdata.MultiSystems:
    """(qmwater-qmwater_not)-(mmwater-mmwater_not)
    = (qmwater-mmwater) - (qmwater_not-mmwater_not)
    """
    ms = dpdata.MultiSystems()
    ep = r'@%EP'
    if cutoff > 0.:
        interactwith = "(%s)<@%f&!%s" % (target, cutoff, ep)
    else:
        interactwith = target

    s_qmwater = dpdata.LabeledSystem(
        "qmwater", fmt='amber/md/qmmm', qm_region=target)
    s_qmwater = s_qmwater.remove_atom_names('EP')
    s_mmwater = dpdata.LabeledSystem(
        "mmwater", fmt='amber/md/qmmm', qm_region=target)
    s_mmwater = s_mmwater.remove_atom_names('EP')
    assert(np.all(s_qmwater['coords']==s_mmwater['coords']))
    s_allcorr = s_mmwater.correction(s_qmwater)
    
    s_qmwaternot = dpdata.LabeledSystem(
        "qmwater_not", fmt='amber/md/qmmm')
    s_qmwaternot = s_qmwaternot.remove_atom_names('EP')
    s_mmwaternot = dpdata.LabeledSystem(
        "mmwater_not", fmt='amber/md/qmmm')
    s_mmwaternot = s_mmwaternot.remove_atom_names('EP')
    assert(np.all(s_qmwaternot['coords']==s_mmwaternot['coords']))
    s_notcorr = s_mmwaternot.correction(s_qmwaternot)

    # manually process
    natom_not = s_notcorr.get_natoms()
    assert(np.all(s_allcorr['coords'][:,-natom_not:]==s_notcorr['coords']))
    s_allcorr.data['energies'] -= s_notcorr['energies']
    s_allcorr.data['forces'][:,-natom_not:] -= s_notcorr['forces']

    s_allcorr = s_allcorr.pick_by_amber_mask(
        "noepw.parm7", interactwith, pass_coords=True, nopbc=True)
    for ss in s_allcorr:
        ms.append(ss)

    if out:
        ms.to_deepmd_npy(out)
    return ms


def run(args):
    get_amber_fp(cutoff=args.cutoff,
                 target=args.qm_region,
                 out=args.out,
                 )



import re
import os
from scipy.io import netcdf
import numpy as np
from dpdata.amber.mask import pick_by_amber_mask
from dpdata.unit import EnergyConversion
from dpdata.periodic_table import ELEMENTS
from dpdata.format import Format
from dpdata.plugins.amber import AmberMDFormat
from ase.geometry import cellpar_to_cell


kcalmol2eV = EnergyConversion("kcal_mol", "eV").value()
symbols = ['X'] + ELEMENTS

energy_convert = kcalmol2eV
force_convert = energy_convert


def read_amber_traj(parm7_file, nc_file, mdfrc_file=None, mden_file=None, mdout_file=None,
                    qm_region=None, labeled=True,
                    ):
    """The amber trajectory includes:
    * nc, NetCDF format, stores coordinates
    * mdfrc, NetCDF format, stores forces
    * mden (optional), text format, stores energies
    * mdout (optional), text format, may store energies if there is no mden_file
    * parm7, text format, stores types

    Parameters
    ----------
    parm7_file, nc_file, mdfrc_file, mden_file, mdout_file: str
      filenames
    qm_region: None or list or str
      amber mask
    """

    flag_atom_type = False
    flag_atom_numb = False
    amber_types = []
    atomic_number = []
    with open(parm7_file) as f:
        for line in f:
            if line.startswith("%FLAG"):
                flag_atom_type = line.startswith("%FLAG AMBER_ATOM_TYPE")
                flag_atom_numb = line.startswith("%FLAG ATOMIC_NUMBER")
            elif flag_atom_type or flag_atom_numb:
                if line.startswith("%FORMAT"):
                    fmt = re.findall(r'\d+', line)
                    fmt0 = int(fmt[0])
                    fmt1 = int(fmt[1])
                else:
                    for ii in range(fmt0):
                        start_index = ii * fmt1
                        end_index = (ii + 1) * fmt1
                        if end_index >= len(line):
                            continue
                        content = line[start_index:end_index].strip()
                        if flag_atom_type:
                            amber_types.append(content)
                        elif flag_atom_numb:
                            atomic_number.append(int(content))
    ml_types = []
    if qm_region is None:
        qm_region = []

    if isinstance(qm_region, str):
        qm_region = pick_by_amber_mask(
            parm7_file, qm_region)
    
    for ii, (tt, nn) in enumerate(zip(amber_types, atomic_number)):
        if ii in qm_region:
            ml_types.append(symbols[nn])
        elif tt in ("OW", "HW", "EP"):
            ml_types.append(tt)
        else:
            ml_types.append("m" + symbols[nn])

    with netcdf.netcdf_file(nc_file, 'r') as f:
        coords = np.array(f.variables["coordinates"][:])
        cell_lengths = np.array(f.variables["cell_lengths"][:])
        cell_angles = np.array(f.variables["cell_angles"][:])
        shape = cell_lengths.shape
        cells = np.zeros((shape[0], 3, 3))
        for ii in range(cell_lengths.shape[0]):
            cells[ii, :, :] = cellpar_to_cell([*cell_lengths[ii], *cell_angles[ii]])

    if labeled:
        with netcdf.netcdf_file(mdfrc_file, 'r') as f:
            forces = np.array(f.variables["forces"][:])

        # load energy from mden_file or mdout_file
        energies = []
        if mden_file is not None and os.path.isfile(mden_file):
            with open(mden_file) as f:
                for line in f:
                    if line.startswith("L6"):
                        s = line.split()
                        if s[2] != "E_pot":
                            energies.append(float(s[2]))
        elif mdout_file is not None and os.path.isfile(mdout_file):
            with open(mdout_file) as f:
                for line in f:
                    if "EPtot" in line:
                        s = line.split()
                        energies.append(float(s[-1]))
        else:
            raise RuntimeError(
                "Please provide one of mden_file and mdout_file")

    atom_names, atom_types, atom_numbs = np.unique(
        ml_types, return_inverse=True, return_counts=True)

    data = {}
    data['atom_names'] = list(atom_names)
    data['atom_numbs'] = list(atom_numbs)
    data['atom_types'] = atom_types
    if labeled:
        data['forces'] = forces * force_convert
        data['energies'] = np.array(energies) * energy_convert
    data['coords'] = coords
    data['cells'] = cells
    data['orig'] = np.array([0, 0, 0])
    return data


@Format.register("amber/md/qmmm")
class AmberMDQMMMFormat(AmberMDFormat):
    def from_system(self, file_name=None, parm7_file=None, nc_file=None, qm_region=None, **kwargs):
        # assume the prefix is the same if the spefic name is not given
        if parm7_file is None:
            parm7_file = file_name + ".parm7"
        if nc_file is None:
            nc_file = file_name + ".nc"
        return read_amber_traj(parm7_file=parm7_file, nc_file=nc_file, qm_region=qm_region, labeled=False)

    def from_labeled_system(self, file_name=None, parm7_file=None, nc_file=None, mdfrc_file=None, mden_file=None, mdout_file=None, qm_region=None, **kwargs):
        # assume the prefix is the same if the spefic name is not given
        if parm7_file is None:
            parm7_file = file_name + ".parm7"
        if nc_file is None:
            nc_file = file_name + ".nc"
        if mdfrc_file is None:
            mdfrc_file = file_name + ".mdfrc"
        if mden_file is None:
            mden_file = file_name + ".mden"
        if mdout_file is None:
            mdout_file = file_name + ".mdout"
        return read_amber_traj(parm7_file, nc_file, mdfrc_file, mden_file, mdout_file, qm_region)



import argparse

from .model_devi import run as model_devi_run
from .corr import run as corr_run
from .qmbuffer import run as qmwater_run
from .qmwater_corr import run as qmwater_corr_run


def run():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()
    parser_model_devi = subparsers.add_parser(
        "devi", help="Calculate model deviations.")
    parser_model_devi.add_argument('--models', type=str, nargs='+',
                                   help="models")
    parser_model_devi.add_argument('--cutoff', type=float,
                                   help="cutoff")
    parser_model_devi.add_argument('--parm7_file', type=str, default="qmmm.parm7",
                                   help="parm7_file")
    parser_model_devi.add_argument('--qm_region', type=str, default=":1",
                                   help="qm_region")
    parser_model_devi.set_defaults(func=model_devi_run)

    parser_corr = subparsers.add_parser(
        "corr", help="Generate systems for DPRc.")
    parser_corr.add_argument('--cutoff', type=float,
                             help="cutoff")
    parser_corr.add_argument('--parm7_file', type=str, default="qmmm.parm7",
                             help="parm7_file")
    parser_corr.add_argument('--qm_region', type=str, default=":1",
                             help="qm_region")
    parser_corr.add_argument('--nc', type=str,
                             help="nc file")
    parser_corr.add_argument('--hl', type=str,
                             help="high level file")
    parser_corr.add_argument('--ll', type=str,
                             help="low level file")
    parser_corr.add_argument('--out', type=str, default="dataset",
                             help="output directory (default is dataset)")
    parser_corr.set_defaults(func=corr_run)

    parser_qmwater = subparsers.add_parser(
        "qmwater", help="Generate QM water parm and mdin file.")
    parser_qmwater.add_argument('--cutoff', type=float,
                             help="cutoff")
    parser_qmwater.add_argument('--parm7_file', type=str, default="qmmm.parm7",
                             help="parm7_file")
    parser_qmwater.add_argument('--qm_region', type=str, default=":1",
                             help="qm_region")
    parser_qmwater.add_argument('--nc', type=str,
                             help="nc file")
    parser_qmwater.add_argument('--hl', type=str,
                             help="high level mdin file")
    parser_qmwater.add_argument('--ll', type=str,
                             help="low level mdin file")
    parser_qmwater.add_argument('--charge', type=str, default="dataset",
                             help="Charge of QM region")
    parser_qmwater.set_defaults(func=qmwater_run)

    parser_qmwater_corr = subparsers.add_parser(
        "qmwater_corr", help="Generate systems for QM water.")
    parser_qmwater_corr.add_argument('--cutoff', type=float,
                             help="cutoff")
    parser_qmwater_corr.add_argument('--qm_region', type=str, default=":1",
                             help="qm_region")
    parser_qmwater_corr.add_argument('--out', type=str, default="dataset",
                             help="output directory (default is dataset)")
    parser_qmwater_corr.set_defaults(func=qmwater_corr_run)

    args = parser.parse_args()
    args.func(args)



"""Calculate model deviation of the Amber MD traj."""
import numpy as np
import dpdata
from dpdata.amber.mask import load_param_file
from tqdm import tqdm


def calculate_devi(models: list,
                   cutoff: float,
                   parm7_file: str = "qmmm.parm7",
                   qm_region: str = ":1"
                   ):
    """
    Parameters
    ----------
    models: ["graph.0.pb", "graph.1.pb"]
    """
    # Deep potentials
    import deepmd.DeepPot as DeepPot
    dps = [DeepPot(mm) for mm in models]

    sy = dpdata.System("rc", parm7_file=parm7_file,
                       fmt="amber/md", use_element_symbols=":1")
    parm7 = load_param_file(parm7_file)
    interactwith = "(%s)<@%f&!%s" % (qm_region, cutoff, r"@%EP")

    stds = []
    for ii, ssy in enumerate(tqdm(sy)):
        ssy = ssy.pick_by_amber_mask(
            parm7, maskstr=interactwith, pass_coords=True)
        ssy = ssy[0]
        ssy = ssy.remove_atom_names("EP")

        results = [ssy.predict(dp).pick_by_amber_mask(
            parm7, qm_region) for dp in dps]
        energies = [dp['energies'] for dp in results]
        energies = np.array(list(energies))
        forces = [dp['forces'] for dp in results]
        forces = np.array(list(forces))
        std_e_min = np.min(np.linalg.norm(np.std(energies, axis=0), axis=-1), axis=-1)
        std_e_ave = np.mean(np.linalg.norm(np.std(energies, axis=0), axis=-1), axis=-1)
        std_e_max = np.max(np.linalg.norm(np.std(energies, axis=0), axis=-1), axis=-1)
        std_f_min = np.min(np.linalg.norm(np.std(forces, axis=0), axis=-1), axis=-1)
        std_f_ave = np.mean(np.linalg.norm(np.std(forces, axis=0), axis=-1), axis=-1)
        std_f_max = np.max(np.linalg.norm(np.std(forces, axis=0), axis=-1), axis=-1)
        stds.append([ii, float(std_e_max), float(std_e_min), float(std_e_ave),  float(std_f_max), float(std_f_min), float(std_f_ave)])
    stds = np.array(stds)
    np.savetxt("model_devi.out", stds, header="step max_devi_e min_devi_e avg_devi_e max_devi_f min_devi_f avg_devi_f")

def run(args):
    calculate_devi(
        models = args.models,
        cutoff = args.cutoff,
        parm7_file = args.parm7_file,
        qm_region = args.qm_region
    )



from setuptools import setup

setup(
    name = "dpamber",
    version = "0.1.5",
    install_requires = [
        "numpy",
        "dpdata[amber]>=0.2.2",
        "deepmd-kit",
        "tqdm",
        "parmed",
        "ase",
        "scipy",
    ],
    entry_points = {
        'console_scripts': ['dpamber=dpamber.cli:run'],
        'dpdata.plugins': [
            'amber_md_qmmm=dpamber.read_amber:AmberMDQMMMFormat'
        ]
    },
    packages = ['dpamber'],
)



import dpdata
from dpdata.amber.mask import load_param_file, pick_by_amber_mask
import parmed.tools as PT
from copy import copy


def qmbuffer(cutoff: float,
              ncfile: str,
              parmfile: str,
              hl_mdinfile: str,
              ll_mdinfile: str,
              target: str = ":1",
              targetcharge: int = -2):
    """
    Assume there is only one frame.
    cutoff = 4.5
    """
    mask_str = "((%s)<:%.1f) & !(%s)" % (target, cutoff, target)
    s = dpdata.System("",
        nc_file=ncfile, parm7_file=parmfile, fmt='amber/md', use_element_symbols=target)
    # systems with nearby waters
    parm = load_param_file(parmfile)
    target_idx = pick_by_amber_mask(parm, target, s['coords'][0])
    nearby_idx = pick_by_amber_mask(parm, mask_str, s['coords'][0])
    parm = reorder(parm, target_idx, nearby_idx)
    # Here we need to generate 4 new parms:
    # 1. reordered, QM water
    # 2. reordered, QM water, no target
    # 3. reordered, MM water
    # 4. reordered, MM water, no target
    # parm 1
    new_nearby_idx = range(len(target_idx), len(target_idx) + len(nearby_idx))
    parm_qm = copy(parm)
    parm_qm, nwater = remove_epw(parm_qm, new_nearby_idx)
    water_mask = "(:2-%d)" % (nwater+1)
    qmmask = "(%s)|(%s)" % (target, water_mask)
    sort_parm(parm_qm)
    write_parm(parm_qm, "qmwater", hl_mdinfile, qmmask, target, charge=targetcharge)

    # parm2
    parm_qm_not = strip_atoms(parm_qm, target)
    water_mask = "(:1-%d)" % (nwater)
    sort_parm(parm_qm_not)
    write_parm(parm_qm_not, "qmwater_not", hl_mdinfile, water_mask, "")

    # parm3
    parm_mm = copy(parm)
    sort_parm(parm_mm)
    write_parm(parm_mm, "mmwater", ll_mdinfile, target, target, charge=targetcharge)

    # parm4
    parm_mm_not = strip_atoms(parm_mm, target)
    sort_parm(parm_mm_not)
    write_parm(parm_mm_not, "mmwater_not", ll_mdinfile, "", "")

    # also try to remove all epw
    parm_noep = copy(parm)
    parm_noep, _ = remove_epw(parm_noep, range(len(target_idx), s.get_natoms()))
    sort_parm(parm_mm)
    parm_noep.write_parm("noepw.parm7")


def label_atoms(parm, target_idx, qmwater_idx):
    """set qwt"""
    for ii in range(len(parm.atoms)):
        atom = parm.atoms[ii]
        if ii in target_idx:
            atom._sorted = 0
        elif ii in qmwater_idx:
            atom._sorted = 1
        else:
            atom._sorted = 2
        

def remove_epw(parm, idx):
    """Remove all EP atoms in water in idx"""
    nwater = 0
    ep = []
    for ii in idx:
        atom = parm.atoms[ii]
        res = atom.residue
        if atom.type == "EP":
            ep.append(ii)
            nwater += 1
            if res.atoms[0].name == "O":
                res.name = "qwt"
                res.atoms[0].charge += atom.charge
                #print(atom.charge, res.atoms[0].charge)
                atom.charge = 0
                # remove from res
                res.atoms.pop(res.atoms.index(atom))
            else:
                raise RuntimeError()
            # remove bonds
            for bond in atom.bonds:
                bond.delete()
            for _ in range(len(atom._bond_partners)):
                atom._bond_partners.pop()
            for _ in range(len(atom._exclusion_partners)):
                atom._exclusion_partners.pop()
        else:
            for p in atom.bonds:
                if p.atom1.type == "EP" or p.atom2.type == "EP":
                    p.delete()
            for _ in range(len(atom._exclusion_partners)):
                atom._exclusion_partners.pop()
            for p in atom.children:
                if p.type == "EP":
                    atom.children.pop(atom.children.index(p))

    mask = "@" + ",".join([ str(ii+1) for ii in ep])
    parm = strip_atoms(parm, mask)
    return parm, nwater

def strip_atoms(parm, mask):
    parm.strip(mask)
    return parm

def reorder(parm, target_idx, nearby_idx):
    label_atoms(parm, target_idx, qmwater_idx=nearby_idx)
    parm.atoms.sort(key=lambda a: a._sorted)
    return parm

def sort_parm(parm):
    # sort residues
    parm.residues.sort(key=sort_residues)

def write_parm(parm, fn, mdinfile, qmmask, target, charge=0):
    parm.write_parm("%s.parm7" % fn)
    parm.write_rst7("%s.rst7" % fn)
    PT.writeCoordinates(parm, '%s.nc' % fn).execute()
    ifqnt = str(int(target!=""))
    with open(mdinfile) as f, open("%s.mdin" % fn, 'w') as fw:
        fw.write(f.read().replace("%QMMASK%", qmmask).replace("%NOSHAKEMASK%", target).replace("%CHARGE%", str(charge)).replace("%IFQNT%", ifqnt))


def sort_residues(res):
    keys = {
        "qwt": 1,
        "WAT": 2,
    }
    return keys.get(res.name, 0)

def run(args):
    qmbuffer(cutoff=args.cutoff,
            parmfile=args.parm7_file,
            ncfile=args.nc,
            ll_mdinfile=args.ll,
            hl_mdinfile=args.hl,
            target=args.qm_region,
            targetcharge=args.charge,
            )
